import dis
import struct
import array
import types
import functools
import weakref
import warnings

try:
	_array_to_bytes = array.array.tobytes
except AttributeError:
	_array_to_bytes = array.array.tostring


class _Bytecode:
	def __init__(self):
		code = (lambda: x if x else y).__code__.co_code
		opcode, oparg = struct.unpack_from('BB', code, 2)

		# Starting with Python 3.6, the bytecode format has changed, using
		# 16-bit words (8-bit opcode + 8-bit argument) for each instruction,
		# as opposed to previously 24 bit (8-bit opcode + 16-bit argument)
		# for instructions that expect an argument and otherwise 8 bit.
		# https://bugs.python.org/issue26647
		if dis.opname[opcode] == 'POP_JUMP_IF_FALSE':
			self.argument = struct.Struct('B')
			self.have_argument = 0
			# As of Python 3.6, jump targets are still addressed by their
			# byte unit. This is matter to change, so that jump targets,
			# in the future might refer to code units (address in bytes / 2).
			# https://bugs.python.org/issue26647
			self.jump_unit = 8 // oparg
		else:
			self.argument = struct.Struct('<H')
			self.have_argument = dis.HAVE_ARGUMENT
			self.jump_unit = 1

		self.has_loop_blocks = 'SETUP_LOOP' in dis.opmap
		self.has_pop_except = 'POP_EXCEPT' in dis.opmap
		self.has_setup_with = 'SETUP_WITH' in dis.opmap
		self.has_setup_except = 'SETUP_EXCEPT' in dis.opmap
		self.has_begin_finally = 'BEGIN_FINALLY' in dis.opmap

	@property
	def argument_bits(self):
		return self.argument.size * 8


_BYTECODE = _Bytecode()


# use a weak dictionary in case code objects can be garbage-collected
_patched_code_cache = weakref.WeakKeyDictionary()
try:
	_patched_code_cache[_Bytecode.__init__.__code__] = None
except TypeError:
	_patched_code_cache = {}  # ...unless not supported


def _make_code(code, codestring):
	try:
		return code.replace(co_code=codestring)  # new in 3.8+
	except AttributeError:
		args = [
			code.co_argcount,  code.co_nlocals,     code.co_stacksize,
			code.co_flags,     codestring,          code.co_consts,
			code.co_names,     code.co_varnames,    code.co_filename,
			code.co_name,      code.co_firstlineno, code.co_lnotab,
			code.co_freevars,  code.co_cellvars
		]

		try:
			args.insert(1, code.co_kwonlyargcount)  # PY3
		except AttributeError:
			pass

		return types.CodeType(*args)


def _parse_instructions(code, yield_nones_at_end=0):
	extended_arg = 0
	extended_arg_offset = None
	pos = 0

	while pos < len(code):
		offset = pos
		if extended_arg_offset is not None:
			offset = extended_arg_offset

		opcode = struct.unpack_from('B', code, pos)[0]
		pos += 1

		oparg = None
		if opcode >= _BYTECODE.have_argument:
			oparg = extended_arg | _BYTECODE.argument.unpack_from(code, pos)[0]
			pos += _BYTECODE.argument.size

			if opcode == dis.EXTENDED_ARG:
				extended_arg = oparg << _BYTECODE.argument_bits
				extended_arg_offset = offset
				continue

		extended_arg = 0
		extended_arg_offset = None
		yield (dis.opname[opcode], oparg, offset)

	for _ in range(yield_nones_at_end):
		yield (None, None, None)


def _get_instruction_size(opname, oparg=0):
	size = 1

	extended_arg = oparg >> _BYTECODE.argument_bits
	if extended_arg != 0:
		size += _get_instruction_size('EXTENDED_ARG', extended_arg)
		oparg &= (1 << _BYTECODE.argument_bits) - 1

	opcode = dis.opmap[opname]
	if opcode >= _BYTECODE.have_argument:
		size += _BYTECODE.argument.size

	return size


def _get_instructions_size(ops):
	size = 0
	for op in ops:
		if isinstance(op, str):
			size += _get_instruction_size(op)
		else:
			size += _get_instruction_size(*op)
	return size


def _write_instruction(buf, pos, opname, oparg=0):
	extended_arg = oparg >> _BYTECODE.argument_bits
	if extended_arg != 0:
		pos = _write_instruction(buf, pos, 'EXTENDED_ARG', extended_arg)
		oparg &= (1 << _BYTECODE.argument_bits) - 1

	opcode = dis.opmap[opname]
	buf[pos] = opcode
	pos += 1

	if opcode >= _BYTECODE.have_argument:
		_BYTECODE.argument.pack_into(buf, pos, oparg)
		pos += _BYTECODE.argument.size

	return pos


def _write_instructions(buf, pos, ops):
	for op in ops:
		if isinstance(op, str):
			pos = _write_instruction(buf, pos, op)
		else:
			pos = _write_instruction(buf, pos, *op)
	return pos


def _warn_bug(msg):
	warnings.warn("Internal error detected" +
				  " - result of with_goto may be incorrect. (%s)" % msg)


class _BlockStack(object):
	def __init__(self, labels, gotos):
		self.stack = []
		self.block_counter = 0
		self.last_block = None
		self.labels = labels
		self.gotos = gotos

	def _replace_in_stack(self, stack, old_block, new_block):
		for i, block in enumerate(stack):
			if block == old_block:
				stack[i] = new_block

	def replace(self, old_block, new_block):
		self._replace_in_stack(self.stack, old_block, new_block)

		for label in self.labels:
			_, _, label_blocks = self.labels[label]
			self._replace_in_stack(label_blocks, old_block, new_block)

		for goto in self.gotos:
			_, _, _, goto_blocks = goto
			self._replace_in_stack(goto_blocks, old_block, new_block)

	def push(self, opname, target_offset=None, previous=None):
		self.block_counter += 1
		self.stack.append((opname, target_offset,
						   previous, self.block_counter))

	def pop(self):
		if self.stack:
			self.last_block = self.stack.pop()
			return self.last_block
		else:
			_warn_bug("can't pop block")

	def pop_of_type(self, type):
		if self.stack and self.top()[0] != type:
			_warn_bug("mismatched block type")
		else:
			return self.pop()

	def copy_to_list(self):
		return list(self.stack)

	def top(self):
		return self.stack[-1] if self.stack else None

	def __len__(self):
		return len(self.stack)


def _find_labels_and_gotos(code):
	labels = {}
	gotos = []

	block_stack = _BlockStack(labels, gotos)

	opname1 = oparg1 = offset1 = None
	opname2 = oparg2 = offset2 = None
	opname3 = oparg3 = offset3 = None

	for opname4, oparg4, offset4 in _parse_instructions(code.co_code, 3):
		endoffset1 = offset2

		# check for block exits
		while block_stack and offset1 == block_stack.top()[1]:
			exit_block = block_stack.pop()
			exit_name = exit_block[0]

			if exit_name == 'SETUP_EXCEPT' and _BYTECODE.has_pop_except:
				block_stack.push('<EXCEPT>', previous=exit_block)
			elif exit_name == 'SETUP_FINALLY':
				block_stack.push('<FINALLY>', previous=exit_block)

		# check for special opcodes
		if opname1 in ('LOAD_GLOBAL', 'LOAD_NAME'):
			if opname2 == 'LOAD_ATTR' and opname3 == 'POP_TOP':
				name = code.co_names[oparg1]
				if name == 'label':
					if oparg2 in labels:
						raise SyntaxError('Ambiguous label {0!r}'.format(
							code.co_names[oparg2]
						))
					labels[oparg2] = (offset1,
									  offset4,
									  block_stack.copy_to_list())
				elif name == 'goto':
					gotos.append((offset1,
								  offset4,
								  oparg2,
								  block_stack.copy_to_list()))
		elif (opname1 in ('SETUP_LOOP',
						  'SETUP_EXCEPT', 'SETUP_FINALLY',
						  'SETUP_WITH', 'SETUP_ASYNC_WITH')) or \
			 (not _BYTECODE.has_loop_blocks and opname1 == 'FOR_ITER'):
			block_stack.push(opname1, endoffset1 + oparg1)
		elif opname1 == 'POP_EXCEPT':
			top_block = block_stack.top()
			if not _BYTECODE.has_setup_except and \
			   top_block and top_block[0] == '<FINALLY>':
				# in 3.8, only finally blocks are supported, so we must
				# determine whether it's except/finally ourselves
				block_stack.replace(top_block,
									('<EXCEPT>',) + top_block[1:])
				_, _, setup_block, _ = top_block
				block_stack.replace(setup_block,
									('SETUP_EXCEPT',) + setup_block[1:])
			block_stack.pop_of_type('<EXCEPT>')
		elif opname1 == 'END_FINALLY':
			# Python puts END_FINALLY at the very end of except
			# clauses, so we must ignore it in the wrong place.
			if block_stack and block_stack.top()[0] == '<FINALLY>':
				block_stack.pop_of_type('<FINALLY>')
		elif opname1 in ('WITH_CLEANUP', 'WITH_CLEANUP_START'):
			if _BYTECODE.has_setup_with:
				# temporary block to match END_FINALLY
				block_stack.push('<FINALLY>')
			else:
				# python 2.6 - finally was actually with
				last_block = block_stack.last_block
				block_stack.replace(last_block,
									('SETUP_WITH',) + last_block[1:])

		opname1, oparg1, offset1 = opname2, oparg2, offset2
		opname2, oparg2, offset2 = opname3, oparg3, offset3
		opname3, oparg3, offset3 = opname4, oparg4, offset4

	if block_stack:
		_warn_bug("block stack not empty")

	return labels, gotos


def _inject_nop_sled(buf, pos, end):
	while pos < end:
		pos = _write_instruction(buf, pos, 'NOP')


def _patch_code(code):
	new_code = _patched_code_cache.get(code)
	if new_code is not None:
		return new_code

	labels, gotos = _find_labels_and_gotos(code)
	buf = array.array('B', code.co_code)

	for pos, end, _ in labels.values():
		_inject_nop_sled(buf, pos, end)

	for pos, end, label, origin_stack in gotos:
		try:
			_, target, target_stack = labels[label]
		except KeyError:
			raise SyntaxError('Unknown label {0!r}'.format(
				code.co_names[label]
			))

		target_depth = len(target_stack)
		if origin_stack[:target_depth] != target_stack:
			raise SyntaxError('Jump into different block')

		ops = []
		for block, _, _, _ in reversed(origin_stack[target_depth:]):
			if block == 'FOR_ITER':
				ops.append('POP_TOP')
			elif block == '<EXCEPT>':
				ops.append('POP_EXCEPT')
			elif block == '<FINALLY>':
				ops.append('END_FINALLY')
			else:
				ops.append('POP_BLOCK')
				if block in ('SETUP_WITH', 'SETUP_ASYNC_WITH'):
					ops.append('POP_TOP')
				# END_FINALLY is needed only in pypy,
				# but seems logical everywhere
				if block in ('SETUP_FINALLY', 'SETUP_WITH',
							 'SETUP_ASYNC_WITH'):
					ops.append('BEGIN_FINALLY' if
							   _BYTECODE.has_begin_finally else
							   ('LOAD_CONST', code.co_consts.index(None)))
					ops.append('END_FINALLY')

		ops.append(('JUMP_ABSOLUTE', target // _BYTECODE.jump_unit))

		if pos + _get_instructions_size(ops) > end:
			# not enough space, add code at buffer end and jump there
			buf_end = len(buf)

			go_to_end_ops = [('JUMP_ABSOLUTE', buf_end // _BYTECODE.jump_unit)]

			if pos + _get_instructions_size(go_to_end_ops) > end:
				# not sure if reachable
				raise SyntaxError('Goto in an incredibly huge function')

			pos = _write_instructions(buf, pos, go_to_end_ops)
			_inject_nop_sled(buf, pos, end)

			buf.extend([0] * _get_instructions_size(ops))
			_write_instructions(buf, buf_end, ops)
		else:
			pos = _write_instructions(buf, pos, ops)
			_inject_nop_sled(buf, pos, end)

	new_code = _make_code(code, _array_to_bytes(buf))

	_patched_code_cache[code] = new_code
	return new_code


def with_goto(func_or_code):
	if isinstance(func_or_code, types.CodeType):
		return _patch_code(func_or_code)

	return functools.update_wrapper(
		types.FunctionType(
			_patch_code(func_or_code.__code__),
			func_or_code.__globals__,
			func_or_code.__name__,
			func_or_code.__defaults__,
			func_or_code.__closure__,
		),
		func_or_code
	)



import argparse
import basically_ti_basic as btb
import os
import tempfile
import sys
import subprocess
import io
import pty

try:
	from .tiParser import TIBasicParser
	from .__version__ import __version__
except ImportError:
	from tiParser import TIBasicParser
	from __version__ import __version__


def isUTF8(file):
	with open(file, 'rb') as f:
		data = f.read()
	try:
		decoded = data.decode('UTF-8')
	except UnicodeDecodeError:
		return False
	else:
		for ch in decoded:
			if 0xD800 <= ord(ch) <= 0xDFFF:
				return False
		return True


def transpile(infile, outfile="stdout", decompileFile=True, forceDecompile=False, multiplication=True, floating_point=True, turbo_draw=False, run=False):

	# detect stdin
	if hasattr(infile, 'name'):
		if not os.path.exists(infile.name):
			# Don't auto close since we are shadowing infile
			temp_stdin = tempfile.NamedTemporaryFile()
			temp_stdin.write(infile.buffer.read())
			temp_stdin.seek(0)
			infile = temp_stdin.name
		else:
			infile = infile.name

	decode = not isUTF8(infile) and decompileFile is True

	if decode is True or forceDecompile is True:
		# Decompile 8Xp file
		with tempfile.NamedTemporaryFile() as f:
			btb.decompile_file(infile, f.name)
			with open(f.name, 'r') as fp:
				pythonCode = TIBasicParser([line.strip() for line in fp.readlines()], multiplication, floating_point, turbo_draw).toPython()

	else:
		# Dont decompile

		# Read infile
		with open(infile, 'r') as f:
			file_lines = [line.strip() for line in f.readlines()]

		pythonCode = TIBasicParser(file_lines, multiplication, floating_point, turbo_draw).toPython()

	# Close temp_stdin if used
	if 'temp_stdin' in vars() or 'temp_stdin' in globals():
		temp_stdin.close()

	# Write to outfile
	if outfile == "stdout":
		if run is False:
			print("\n".join(pythonCode))
		else:
			with tempfile.NamedTemporaryFile() as f:
				for line in pythonCode:
					f.write(line.encode() + b"\n")
				f.seek(0)

				# test if something was piped to stdin
				if 'temp_stdin' in vars() or 'temp_stdin' in globals():
					# Create new tty to handle ioctl errors in termios
					master_fd, slave_fd = pty.openpty()
					proc = subprocess.Popen([sys.executable, f.name], stdin=slave_fd)
				else:
					# Nothing was piped to stdin
					proc = subprocess.Popen([sys.executable, f.name])

				try:
					proc.wait()
				except (Exception, KeyboardInterrupt):
					proc.terminate()
	else:
		with open(outfile, 'w') as f:
			for line in pythonCode:
				f.write(line + "\n")
		if run is True:
			os.system(sys.executable + " " + outfile)


def main():
	parser = argparse.ArgumentParser(description='TI-BASIC to Python 3 Transpiler')
	infile_argument = parser.add_argument(
		'infile',
		metavar='infile',
		nargs='?',
		type=argparse.FileType('r'),
		default=(None if sys.stdin.isatty() else sys.stdin),
		help="Input file (filename or stdin)."
	)
	parser.add_argument(
		'-o',
		'--out',
		required=False,
		default='stdout',
		dest='outfile',
		help="Optional output file to write to. Defaults to standard out."
	)
	parser.add_argument(
		'-n',
		'--force-normal',
		action='store_false',
		help="Forces the program to not attempt and decompile the input file. Useful for false positives",
		dest="n"
	)
	parser.add_argument(
		'-d',
		'--force-decompile',
		action="store_true",
		help="Forces the program to attempt to decompile the input file",
		dest="d"
	)

	parser.add_argument(
		'--no-fix-multiplication',
		action='store_false',
		help='Do not attempt to fix implicit multiplication. For example, AB -> A*B and A(1) -> A*(1)',
		dest='multiplication'
	)

	parser.add_argument(
		'--no-fix-floating-point',
		action='store_false',
		help='Do not attempt to fix floating point arithmetic errors. For example, 1.1 * 3 would normally say 3.3000000000000003 instead of 3.3',
		dest='floating_point'
	)

	parser.add_argument(
		'--turbo-draw',
		action='store_true',
		help='Remove the 0.1 second delay between drawing actions'
	)

	parser.add_argument(
		'-r',
		'--run',
		action="store_true",
		help="Runs the program after it\'s done transpiling. Will not print to stdout",
		dest='run'
	)

	parser.add_argument(
		'-V',
		'--version',
		action='version',
		version='ti842py {version}'.format(version=__version__)
	)

	args = parser.parse_args()

	if hasattr(args.infile, '__getitem__'):
		infile = args.infile[0]
	else:
		infile = args.infile

	if infile is None:
		raise argparse.ArgumentError(infile_argument, 'the infile argument is required')

	transpile(infile, args.outfile, args.n, args.d, args.multiplication, args.floating_point, args.turbo_draw, args.run)


if __name__ == "__main__":
	main()



import pynput

class GetKey:
	def __init__(self):
		self.last_key = 0
		self.keymap = {
			pynput.keyboard.Key.f1: 11,
			pynput.keyboard.Key.f2: 12,
			pynput.keyboard.Key.f3: 13,
			pynput.keyboard.Key.f4: 14,
			pynput.keyboard.Key.f5: 15,
			"`": 21,
			pynput.keyboard.Key.f6: 22,
			pynput.keyboard.Key.delete: 23,
			pynput.keyboard.Key.left: 24,
			pynput.keyboard.Key.up: 25,
			pynput.keyboard.Key.right: 26,
			"~": 31,
			pynput.keyboard.Key.f10: 32,
			pynput.keyboard.Key.f7: 33,
			pynput.keyboard.Key.down: 34,
			"a": 41,
			"b": 42,
			"c": 43,
			pynput.keyboard.Key.f8: 44,
			pynput.keyboard.Key.f9: 45,
			"d": 51,
			"e": 52,
			"f": 53,
			"g": 54,
			"h": 55,
			"^": 55,
			"i": 61,
			",": 62,
			"j": 62,
			"(": 63,
			"k": 63,
			")": 64,
			"l": 64,
			"/": 65,
			"m": 65,
			"÷": 65,
			"n": 71,
			"7": 72,
			"o": 72,
			"8": 73,
			"p": 73,
			"9": 74,
			"q": 74,
			"×": 75,
			"*": 75,
			"r": 75,
			"s": 81,
			"4": 82,
			"t": 82,
			"5": 83,
			"u": 83,
			"6": 84,
			"v": 84,
			"-": 85,
			"−": 85,
			"w": 85,
			"=": 91,
			"x": 91,
			"1": 92,
			"y": 92,
			"2": 93,
			"z": 93,
			"3": 94,
			"+": 95,
			"\"": 95,
			"0": 102,
			" ": 102,
			".": 103,
			":": 103,
			"?": 104,
			pynput.keyboard.Key.enter: 105
		}

	def set_last_key(self, key):
		if hasattr(key, 'char'):
			# convert to string
			key = key.char.lower()

		if key in self.keymap:
			self.last_key = self.keymap[key]

	def get_last_key(self):
		key = self.last_key
		self.last_key = 0
		return key

get_key = GetKey()
listener = pynput.keyboard.Listener(on_press=get_key.set_last_key)
listener.start()



"""
MIT License

Copyright (c) 2020 André Roberge

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""token_utils.py
------------------

A collection of useful functions and methods to deal with tokenizing
source code.
"""
import ast
import keyword
import tokenize as py_tokenize

from io import StringIO

__version__ = "0.1.5"
_token_format = "type={type}  string={string}  start={start}  end={end}  line={line}"


class Token:
	"""Token as generated from Python's tokenize.generate_tokens written here in
	a more convenient form, and with some custom methods.

	The various parameters are::

		type: token type
		string: the token written as a string
		start = (start_row, start_col)
		end = (end_row, end_col)
		line: entire line of code where the token is found.

	Token instances are mutable objects. Therefore, given a list of tokens,
	we can change the value of any token's attribute, untokenize the list and
	automatically obtain a transformed source.
	"""

	def __init__(self, token):
		self.type = token[0]
		self.string = token[1]
		self.start = self.start_row, self.start_col = token[2]
		self.end = self.end_row, self.end_col = token[3]
		self.line = token[4]

	def __eq__(self, other):
		"""Compares a Token with another object; returns true if
		self.string == other.string or if self.string == other.
		"""
		if hasattr(other, "string"):
			return self.string == other.string
		elif isinstance(other, str):
			return self.string == other
		else:
			raise TypeError(
				"A token can only be compared to another token or to a string."
			)

	def __repr__(self):
		"""Nicely formatted token to help with debugging session.

		Note that it does **not** print a string representation that could be
		used to create a new ``Token`` instance, which is something you should
		never need to do other than indirectly by using the functions
		provided in this module.
		"""
		return _token_format.format(
			type="%s (%s)" % (self.type, py_tokenize.tok_name[self.type]),
			string=repr(self.string),
			start=str(self.start),
			end=str(self.end),
			line=repr(self.line),
		)

	def __str__(self):
		"""Returns the string attribute."""
		return self.string

	def is_comment(self):
		"""Returns True if the token is a comment."""
		return self.type == py_tokenize.COMMENT

	def is_identifier(self):
		"""Returns ``True`` if the token represents a valid Python identifier
		excluding Python keywords.

		Note: this is different from Python's string method ``isidentifier``
		which also returns ``True`` if the string is a keyword.
		"""
		return self.string.isidentifier() and not self.is_keyword()

	def is_name(self):
		"""Returns ``True`` if the token is a type NAME"""
		return self.type == py_tokenize.NAME

	def is_keyword(self):
		"""Returns True if the token represents a Python keyword."""
		return keyword.iskeyword(self.string)

	def is_number(self):
		"""Returns True if the token represents a number"""
		return self.type == py_tokenize.NUMBER

	def is_float(self):
		"""Returns True if the token represents a float"""
		return self.is_number() and isinstance(ast.literal_eval(self.string), float)

	def is_integer(self):
		"""Returns True if the token represents an integer"""
		return self.is_number() and isinstance(ast.literal_eval(self.string), int)

	def is_complex(self):
		"""Returns True if the token represents a complex number"""
		return self.is_number() and isinstance(ast.literal_eval(self.string), complex)

	def is_space(self):
		"""Returns True if the token indicates a change in indentation,
		the end of a line, or the end of the source
		(``INDENT``, ``DEDENT``, ``NEWLINE``, ``NL``, and ``ENDMARKER``).

		Note that spaces, including tab charcters ``\\t``, between tokens
		on a given line are not considered to be tokens themselves.
		"""
		return self.type in (
			py_tokenize.INDENT,
			py_tokenize.DEDENT,
			py_tokenize.NEWLINE,
			py_tokenize.NL,
			py_tokenize.ENDMARKER,
		)

	def is_string(self):
		"""Returns True if the token is a string"""
		return self.type == py_tokenize.STRING

	def is_in(self, iterable):
		"""Returns True if the string attribute is found as an item of iterable."""
		return self.string in iterable

	def is_not_in(self, iterable):
		"""Returns True if the string attribute is found as an item of iterable."""
		return self.string not in iterable


def find_token_by_position(tokens, row, column):
	"""Given a list of tokens, a specific row (linenumber) and column,
	a two-tuple is returned that includes the token
	found at that position as well as its list index.

	If no such token can be found, ``None, None`` is returned.
	"""
	for index, tok in enumerate(tokens):
		if (
			tok.start_row <= row <= tok.end_row
			and tok.start_col <= column < tok.end_col
		):
			return tok, index
	return None, None


def fix_empty_line(source, tokens):
	"""Python's tokenizer drops entirely a last line if it consists only of
	space characters and/or tab characters.  To ensure that we can always have::

		untokenize(tokenize(source)) == source

	we correct the last token content if needed.
	"""
	nb = 0
	for char in reversed(source):
		if char in (" ", "\t"):
			nb += 1
		else:
			break
	tokens[-1].string = source[-nb:]


def tokenize(source, warning=True):
	"""Transforms a source (string) into a list of Tokens.

	If an exception is raised by Python's tokenize module, the list of tokens
	accumulated up to that point is returned.
	"""
	tokens = []

	for tok in py_tokenize.generate_tokens(StringIO(source).readline):
		try:
			token = Token(tok)
			tokens.append(token)
		except (py_tokenize.TokenError, Exception) as exc:
			if warning:
				print(
					"WARNING: the following error was raised in ",
					f"{__name__}.tokenize",
				)
				print(exc)
			return tokens

	if source.endswith((" ", "\t")):
		fix_empty_line(source, tokens)

	return tokens


def get_significant_tokens(source):
	"""Gets a list of tokens from a source (str), ignoring comments
	as well as any token whose string value is either null or
	consists of spaces, newline or tab characters.

	If an exception is raised by Python's tokenize module, the list of tokens
	accumulated up to that point is returned.
	"""
	tokens = []
	try:
		for tok in py_tokenize.generate_tokens(StringIO(source).readline):
			token = Token(tok)
			if not token.string.strip():
				continue
			if token.is_comment():
				continue
			tokens.append(token)
	except py_tokenize.TokenError:
		return tokens

	return tokens


def get_lines(source):
	"""Transforms a source (string) into a list of Tokens, with each
	(inner) list containing all the tokens found on a given line of code.
	"""
	lines = []
	current_row = -1
	new_line = []
	for tok in py_tokenize.generate_tokens(StringIO(source).readline):
		try:
			token = Token(tok)
			if token.start_row != current_row:
				current_row = token.start_row
				if new_line:
					lines.append(new_line)
				new_line = []
			new_line.append(token)
		except (py_tokenize.TokenError, Exception) as exc:
			print(
				"WARNING: the following tokenize error was raised in "
				f"{__name__}.get_lines"
			)
			print(exc)

	if new_line:
		lines.append(new_line)
	if source.endswith((" ", "\t")):
		fix_empty_line(source, lines[-1])
	return lines


def get_number(tokens, exclude_comment=True):
	"""Given a list of tokens, gives a count of the number of
	tokens which are not space tokens (such as ``NEWLINE``, ``INDENT``,
	``DEDENT``, etc.)

	By default, ``COMMMENT`` tokens are not included in the count.
	If you wish to include them, set ``exclude_comment`` to ``False``.
	"""
	nb = len(tokens)
	for token in tokens:
		if token.is_space():
			nb -= 1
		elif exclude_comment and token.is_comment():
			nb -= 1
	return nb


def strip_comment(line):
	"""Removes comments from a line"""
	tokens = []
	try:
		for tok in py_tokenize.generate_tokens(StringIO(line).readline):
			token = Token(tok)
			if token.is_comment():
				continue
			tokens.append(token)
	except py_tokenize.TokenError:
		pass
	return untokenize(tokens)


# TODO: add unit test for this
def find_substring_index(main, substring):
	"""Somewhat similar to the find() method for strings,
	this function determines if the tokens for substring appear
	as a subsequence of the tokens for main. If so, the index
	of the first token in returned, otherwise -1 is returned.
	"""
	main_tokens = [tok.string for tok in get_significant_tokens(main)]
	sub_tokens = [tok.string for tok in get_significant_tokens(substring)]
	for index, token in enumerate(main_tokens):
		if (
			token == sub_tokens[0]
			and main_tokens[index : index + len(sub_tokens)] == sub_tokens
		):
			return index
	return -1


def get_first(tokens, exclude_comment=True):
	"""Given a list of tokens, find the first token which is not a space token
	(such as a ``NEWLINE``, ``INDENT``, ``DEDENT``, etc.) and,
	by default, also not a ``COMMMENT``.

	``COMMMENT`` tokens can be included by setting ``exclude_comment`` to ``False``.

	Returns ``None`` if none is found.
	"""
	for token in tokens:
		if token.is_space() or (exclude_comment and token.is_comment()):
			continue
		return token
	return None


def get_first_index(tokens, exclude_comment=True):
	"""Given a list of tokens, find the index of the first token which is
	not a space token (such as a ``NEWLINE``, ``INDENT``, ``DEDENT``, etc.) nor
	a ``COMMMENT``. If it is desired to include COMMENT, set ``exclude_comment``
	to ``True``.

	Returns ``None`` if none is found.
	"""
	for index, token in enumerate(tokens):
		if token.is_space() or (exclude_comment and token.is_comment()):
			continue
		return index
	return None


def get_last(tokens, exclude_comment=True):
	"""Given a list of tokens, find the last token which is not a space token
	(such as a ``NEWLINE``, ``INDENT``, ``DEDENT``, etc.) and, by default,
	also not a ``COMMMENT``.

	``COMMMENT`` tokens can be included by setting``exclude_comment``
	to ``False``.

	Returns ``None`` if none is found.
	"""
	return get_first(reversed(tokens), exclude_comment=exclude_comment)


def get_last_index(tokens, exclude_comment=True):
	"""Given a list of tokens, find the index of the last token which is
	not a space token (such as a ``NEWLINE``, ``INDENT``, ``DEDENT``, etc.) nor
	a ``COMMMENT``. If it is desired to include COMMENT, set ``exclude_comment``
	to True.

	Returns ``None`` if none is found.
	"""
	return (
		len(tokens)
		- 1
		- get_first_index(reversed(tokens), exclude_comment=exclude_comment)
	)


def dedent(tokens, nb):
	"""Given a list of tokens, produces an equivalent list corresponding
	to a line of code with the first nb characters removed.
	"""
	line = untokenize(tokens)
	line = line[nb:]
	return tokenize(line)


def indent(tokens, nb, tab=False):
	"""Given a list of tokens, produces an equivalent list corresponding
	to a line of code with nb space characters inserted at the beginning.

	If ``tab`` is specified to be ``True``, ``nb`` tab characters are inserted
	instead of spaces.
	"""
	line = untokenize(tokens)
	if tab:
		line = "\t" * nb + line
	else:
		line = " " * nb + line
	return tokenize(line)


def untokenize(tokens):
	"""Return source code based on tokens.

	Adapted from https://github.com/myint/untokenize,
	Copyright (C) 2013-2018 Steven Myint, MIT License (same as this project).

	This is similar to Python's own tokenize.untokenize(), except that it
	preserves spacing between tokens, by using the line
	information recorded by Python's tokenize.generate_tokens.
	As a result, if the original soure code had multiple spaces between
	some tokens or if escaped newlines were used or if tab characters
	were present in the original source, those will also be present
	in the source code produced by untokenize.

	Thus ``source == untokenize(tokenize(source))``.

	Note: if you you modifying tokens from an original source:

	Instead of full token object, ``untokenize`` will accept simple
	strings; however, it will only insert them *as is* without taking them
	into account when it comes with figuring out spacing between tokens.
	"""
	words = []
	previous_line = ""
	last_row = 0
	last_column = -1
	last_non_whitespace_token_type = None

	for token in tokens:
		if isinstance(token, str):
			words.append(token)
			continue
		if token.type == py_tokenize.ENCODING:
			continue

		# Preserve escaped newlines.
		if (
			last_non_whitespace_token_type != py_tokenize.COMMENT
			and token.start_row > last_row
		):
			if previous_line.endswith(("\\\n", "\\\r\n", "\\\r")):
				words.append(previous_line[len(previous_line.rstrip(" \t\n\r\\")) :])

		# Preserve spacing.
		if token.start_row > last_row:
			last_column = 0
		if token.start_col > last_column:
			words.append(token.line[last_column : token.start_col])

		words.append(token.string)

		previous_line = token.line
		last_row = token.end_row
		last_column = token.end_col
		if not token.is_space():
			last_non_whitespace_token_type = token.type

	return "".join(words)


def print_tokens(source):
	"""Prints tokens found in source, excluding spaces and comments.

	``source`` is either a string to be tokenized, or a list of Token objects.

	This is occasionally useful as a debugging tool.
	"""
	if isinstance(source[0], Token):
		source = untokenize(source)

	for lines in get_lines(source):
		for token in lines:
			print(repr(token))
		print()



import decimal
import copy
import operator
import math

class DimensionMismatchError(Exception):
	pass

class InvalidDimensionError(Exception):
	pass

class SingularMatrixError(Exception):
	pass

class Matrix:
	'''
	Matrix class that acts like the matrices on TI calculators
	'''
	def __init__(self, init_matrix=None):
		if init_matrix is None:
			self.matrix = [[0]]
		else:
			self.matrix = init_matrix

	def reshape(self, rows, cols):
		'''
		Reshape Matrix
		'''

		# No checks required since here aren't any rules to resizing matrices
		# in TI-BASIC for the 84

		result = [[0 for _ in range(cols)] for _ in range(rows)]

		count = 0
		for i in range(len(self.matrix)):
			for j in range(len(self.matrix[0])):
				try:
					result[count // cols][count % cols] = self.matrix[i][j]
				except IndexError:
					# Shrinking the matrix will throw an error; ignore this
					pass

				count += 1

		self.matrix = result

	def shape(self):
		'''
		returns rows, cols
		'''

		return len(self.matrix), len(self.matrix[0])

	def _operate_on_matrix_elements(self, operation, factor_a, factor_b):
		'''
		Apply an operation to each element in the matrix.
		factors can be ints, floats, or other matrices
		'''

		# Convert to decimal objects
		if isinstance(factor_a, float):
			factor_a = decimal.Decimal(str(factor_a))
		if isinstance(factor_b, float):
			factor_b = decimal.Decimal(str(factor_b))

		if isinstance(factor_a, Matrix) and isinstance(factor_b, Matrix):
			if not factor_a.shape() == factor_b.shape():
				raise DimensionMismatchError('Check size (dim) of lists or matrices')

		new_matrix = Matrix()
		if isinstance(factor_a, Matrix):
			new_matrix.reshape(*factor_a.shape())
		elif isinstance(factor_b, Matrix):
			new_matrix.reshape(*factor_b.shape())

		if isinstance(factor_a, Matrix):
			factor_a = copy.deepcopy(factor_a.matrix)
		if isinstance(factor_b, Matrix):
			factor_b = copy.deepcopy(factor_b.matrix)

		for i in range(len(new_matrix)):
			for j in range(len(new_matrix[i])):

				# Generate factors to compute
				if isinstance(factor_a, list):
					item_a = factor_a[i][j]
				else:
					item_a = factor_a

				if isinstance(factor_b, list):
					item_b = factor_b[i][j]
				else:
					item_b = factor_b

				# Convert current items to decimal objects if applicable
				if isinstance(item_a, float):
					item_a = decimal.Decimal(str(item_a))
				if isinstance(item_b, float):
					item_b = decimal.Decimal(str(item_b))


				item_computation = operation(item_a, item_b)
				if isinstance(item_computation, decimal.Decimal):
					if str(item_computation).isdigit():
						item_computation = int(item_computation)
					else:
						item_computation = float(item_computation)

				new_matrix[i][j] = item_computation

		return new_matrix

	def _generate_zero_matrix(self, size):
		'''
		Generate a matrix of shape rows, cols with all zeros, where rows and cols are both equal to size.
		The main diagonal is one.
		Used for raising a matrix to the power of zero.
		'''

		return Matrix([[1 if i == j else 0 for j in range(size)] for i in range(size)])

	def _invert_matrix(self, matrix):
		'''
		Invert a matrix object
		Mostly taken from https://stackoverflow.com/a/39881366/11591238
		'''

		def transposeMatrix(m):
			return map(list, zip(*m))

		def getMatrixMinor(m, i, j):
			return [row[:j] + row[j + 1:] for row in (m[:i] + m[i + 1:])]

		def getMatrixDeternminant(m):
			#base case for 2x2 matrix
			if len(m) == 2:
				return m[0][0] * m[1][1] - m[0][1] * m[1][0]

			determinant = 0
			for c in range(len(m)):
				determinant += ((-1) ** c) * m[0][c] * getMatrixDeternminant(getMatrixMinor(m, 0, c))
			return determinant

		def getMatrixInverse(m):
			determinant = getMatrixDeternminant(m)
			#special case for 2x2 matrix:
			if len(m) == 2:
				return [[m[1][1] / determinant, -1 * m[0][1] / determinant],
						[-1 * m[1][0] / determinant, m[0][0] / determinant]]

			#find matrix of cofactors
			cofactors = []
			for r in range(len(m)):
				cofactorRow = []
				for c in range(len(m)):
					minor = getMatrixMinor(m, r, c)
					cofactorRow.append(((-1) ** (r + c)) * getMatrixDeternminant(minor))
				cofactors.append(cofactorRow)
			cofactors = transposeMatrix(cofactors)
			for r in range(len(cofactors)):
				for c in range(len(cofactors)):
					cofactors[r][c] = cofactors[r][c] / determinant
			return cofactors

		try:
			new_matrix = getMatrixInverse(matrix.matrix)
		except ZeroDivisionError as exc:
			raise SingularMatrixError('No inverse matrix exists.') from exc

		return Matrix(new_matrix)


	def __len__(self):
		# Return rows, cols
		return self.shape()

	def __mul__(self, other):
		if isinstance(other, Matrix):
			# matrix multiplication

			# https://www.varsitytutors.com/hotmath/hotmath_help/topics/compatible-matrices.html
			if self.shape()[1] == other.shape()[0]:

				# https://stackoverflow.com/a/10508239/11591238
				zip_other_matrix = list(zip(*other.matrix))
				new_matrix = [
					[
						sum(ele_a * ele_b for ele_a, ele_b in zip(row_self, col_other))
						for col_other in zip_other_matrix
					] for row_self in self.matrix
				]

				return Matrix(new_matrix)

			else:

				raise DimensionMismatchError('Check size (dim) of lists or matrices')

		elif isinstance(other, (int, float)):
			new_matrix = self._operate_on_matrix_elements(operator.mul, self, other)
			return new_matrix

	def __imul__(self, other):
		new_matrix = self.__mul__(other)
		self.matrix = copy.deepcopy(new_matrix.matrix)
		del new_matrix

		return self

	def __rmul__(self, other):
		new_matrix = self._operate_on_matrix_elements(operator.mul, other, self)
		return new_matrix

	def __add__(self, other):
		# Plus(+) and minus(-) add or subtract the matching elements of two matrices with the same size
		# http://tibasicdev.wikidot.com/matrices
		new_matrix = self._operate_on_matrix_elements(operator.add, self, other)
		return new_matrix

	def __iadd__(self, other):
		new_matrix = self.__add__(other)
		self.matrix = copy.deepcopy(new_matrix.matrix)
		del new_matrix

		return self

	def __radd__(self, other):
		new_matrix = self._operate_on_matrix_elements(operator.add, other, self)
		return new_matrix

	def __sub__(self, other):
		new_matrix = self._operate_on_matrix_elements(operator.sub, self, other)
		return new_matrix

	def __isub__(self, other):
		new_matrix = self.__sub__(other)
		self.matrix = copy.deepcopy(new_matrix.matrix)
		del new_matrix

		return self

	def __rsub__(self, other):
		new_matrix = self._operate_on_matrix_elements(operator.sub, other, self)
		return new_matrix

	def __pow__(self, other):
		if self.shape()[0] != self.shape()[1]:
			raise InvalidDimensionError('Matrix must be square.')

		if other >= 1:
			new_matrix = Matrix(copy.deepcopy(self.matrix))
			for _ in range(other - 1):
				new_matrix = self.__mul__(new_matrix)

			return new_matrix
		elif other == 0:
			return self._generate_zero_matrix(self.shape()[0])
		elif other == -1:
			return self._invert_matrix(self)
		else:
			raise ValueError('Must not be under -1')

	def __ipow__(self, other):
		new_matrix = self.__pow__(other)
		self.matrix = copy.deepcopy(new_matrix.matrix)
		del new_matrix

		return self

	def __floor__(self):
		new_matrix = copy.deepcopy(self.matrix)
		for i in range(len(new_matrix)):
			for j in range(len(new_matrix[i])):
				new_matrix[i][j] = math.floor(new_matrix[i][j])
		return Matrix(new_matrix)

	def __abs__(self):
		new_matrix = copy.deepcopy(self.matrix)
		for i in range(len(new_matrix)):
			for j in range(len(new_matrix[i])):
				new_matrix[i][j] = abs(new_matrix[i][j])
		return Matrix(new_matrix)

	def __neg__(self):
		new_matrix = copy.deepcopy(self.matrix)
		for i in range(len(new_matrix)):
			for j in range(len(new_matrix[i])):
				new_matrix[i][j] = -new_matrix[i][j]
		return Matrix(new_matrix)

	def __eq__(self, other):
		return 1 if self.matrix == other.matrix else 0

	def __ne__(self, other):
		return not self.__eq__(other)

	def __repr__(self):
		return '\n'.join(['[' + ' '.join([str(num) for num in sublist]) + ']' for sublist in self.matrix])

	def __getitem__(self, index):
		# We do not need __setitem__ because the user should not be messing
		# with the parent lists directly. __getitem__ is enough to interface
		# with the child lists contained within self.matrix
		return self.matrix[index] # indexes start at 1 in TI-BASIC

	def __call__(self, row, col):
		# TI-BASIC matrices are accessed like [A](1, 2), so I've implemented
		# that here. Uses __getitem__
		return self.__getitem__(row)[col] # indexes start at 1 in TI-BASIC



import setuptools
import os

install_requires = [
	"basically-ti-basic>=0.1.6",
	"pythondialog>=3.5.1",
	"graphics.py>=5.0.0",
	"pynput>=1.7.3"
]

with open("README.md", "r") as fh:
	long_description = fh.read()

here = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(here, "ti842py", "__version__.py"), "r") as f:
	exec(f.read(), about)

setuptools.setup(
	name=about["__title__"],
	version=about["__version__"],
	author=about["__author__"],
	author_email=about["__author_email__"],
	description=about["__description__"],
	long_description=long_description,
	long_description_content_type="text/markdown",
	url=about["__url__"],
	install_requires=install_requires,
	entry_points={
		'console_scripts': [
			'ti842py = ti842py.main:main'
		]
	},
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Operating System :: OS Independent",
		"Development Status :: 4 - Beta",
		"Intended Audience :: End Users/Desktop",
		"Intended Audience :: Developers"
	],
	python_requires='>=3.6',
	include_package_data=True
)



import decimal
import inspect
import io
import tokenize


def decistmt(s):
	# Function for parsing a line of code and making all floats into decimal objects

	# List containing all tokens of new statement
	result = []

	# Tokenize statement
	tokens = tokenize.tokenize(io.BytesIO(s.encode('utf-8')).readline)

	# Iterate over tokens
	for toknum, tokval, _, _, _ in tokens:

		# Test if current token is a number and contains a '.' in order to detect floats
		if toknum == tokenize.NUMBER and '.' in tokval: # is float

			# Current token is float, so insert a new decimal object into the new statement instead of the current token
			# Example: 3.3 -> decimal.Decimal("3.3")
			result.extend([
				(tokenize.NAME, 'decimal'),
				(tokenize.OP, '.'),
				(tokenize.NAME, 'Decimal'),
				(tokenize.OP, '('),
				(tokenize.STRING, repr(tokval)),
				(tokenize.OP, ')')
			])
		else:
			# Current token is not a float, append it to the new statement without modification
			result.append((toknum, tokval))

	# Convert token list back into source code
	return tokenize.untokenize(result).decode('utf-8')


def fix_floating_point(f):
	# Decorator for fixing floating-point arithmetic within a function
	def wrapper(*args, **kwargs):
		# Get the source of the function as a list (each element is one line)
		src = inspect.getsource(f).split('\n')
		# Initiate variable for creation of new source code. Starts with function definition
		newsrc = src[1] + '\n'

		# Iterate over function source; skip decorator and function definition
		for line in src[2:]:
			# Expand tabs for spaces in order to calculate indent size
			line = line.expandtabs(4)

			# Calculate amount of indentation for line
			indent = len(line) - len(line.lstrip())

			# Parse and adjust line of code with decistmt function defined above, then add to new source code
			newsrc += '{}{}\n'.format(' ' * indent, decistmt(line))

		# Exec new source code after adjustments to create a new local function
		exec(newsrc)

		# Call that new function that we just created; has the same name as the decorated function so we can just use f.__name__
		locals()[f.__name__]()

	# Return wrapper function
	return wrapper


from graphics import *
import time
import tkinter.font as tkfont
import decimal
import re


BLUE = 10
RED = 11
BLACK = 12
MAGENTA = 13
GREEN = 14
ORANGE = 15
BROWN = 16
NAVY = 17
LTBLUE = 18
YELLOW = 19
WHITE = 20
LTGRAY = 21
MEDGRAY = 22
GRAY = 23
DARKGRAY = 24

class InvalidColorError(Exception):
	pass

class Draw:
	def __init__(self):
		self.win = None
		self.winOpen = False
		self.pixels = {}
		self.points = {}
		self.texts = {}
		self.colors = {'10': 'blue', '11': 'red', '12': 'black', '13': 'magenta', '14': 'green', '15': 'orange', '16': 'brown', '17': 'navy', '18': 'light sky blue', '19': 'yellow', '20': 'white', '0': 'white', '21': 'light gray', '22': 'dark gray', '23': 'gray', '24': 'dark slate gray'}
		for _ in range(1, 10):
			self.colors[str(_)] = 'blue'
		self.currentTextColor = 'blue'

	def _slow(function):
		def wrapper(*args, **kwargs):
			function(*args, **kwargs)
			time.sleep(0.01)
		return wrapper

	def openWindow(self):
		# The TI-84 Plus CE has a graph resolution of 250x160 pixels
		# Not to be confused with the screen resolution of 320x240 pixels
		if self.winOpen is False:
			self.win = GraphWin('ti842py', 250, 160)
			self.win.setBackground('white')
			self.winOpen = True

	def closeWindow(self):
		if self.winOpen is True:
			self.win.close()
			self.winOpen = False

	def graphCoordsToPixels(self, x, y, minX=-10, maxX=10, minY=-10, maxY=10):
		# Can work for vertical or horizontal coords
		xs = (decimal.Decimal(maxX) - decimal.Decimal(minX)) / 250
		ys = (decimal.Decimal(minY) - decimal.Decimal(maxY)) / 160
		horiz = (decimal.Decimal(x) - decimal.Decimal(minX)) / decimal.Decimal(xs)
		vertical = (decimal.Decimal(y) - decimal.Decimal(maxY)) / decimal.Decimal(ys)
		return horiz, vertical

	def tiColorToGraphicsColor(self, color, isBackground=False):
		if color not in range(1, 25):
			raise InvalidColorError(f'The specified color value "{color}" is not in range 1-24')
		color = str(color)

		color = re.sub(r'(\d+)', lambda m: self.colors.get(m.group(), '[ERR:DOMAIN]'), color)

		return color

	@_slow
	def backgroundOn(self, color):
		self.win.setBackground(self.tiColorToGraphicsColor(color, isBackground=True))

	@_slow
	def backgroundOff(self):
		self.win.setBackground('white')

	def clrDraw(self):
		for item in self.win.items[:]:
			item.undraw()
		self.win.update()

	@_slow
	def circle(self, x, y, r, color='blue', linestyle=''):
		# TODO: Implement linestyle
		c = Circle(Point(self.graphCoordsToPixels(x), self.graphCoordsToPixels(y)), r)
		c.setOutline(self.tiColorToGraphicsColor(color))
		c.draw(self.win)

	@_slow
	def line(self, x1, y1, x2, y2, erase=1, color='blue', style=1):
		# TODO: Erase and style not implemented yet
		x1, y1 = self.graphCoordsToPixels(x1, y1)
		x2, y2 = self.graphCoordsToPixels(x2, y2)
		line = Line(Point(x1, y1), Point(x2, y2))
		line.setOutline(self.tiColorToGraphicsColor(color))
		line.draw(self.win)

	@_slow
	def textColor(self, color):
		self.currentTextColor = self.tiColorToGraphicsColor(color)

	@_slow
	def text(self, row, column, *args):
		# column = x, row = y
		text = ''.join([str(arg) for arg in args])
		fontsize = 7

		# Calculate correct center value based on text width
		font = tkfont.Font(family='helvetica', size=fontsize, weight='normal')
		text_width = font.measure(text) // 2
		text_height = font.metrics('linespace') // 2
		column += text_width
		row += text_height

		# Undraw previous text
		# TODO: draw background behind it instead, maybe
		if str(column) in self.texts:
			if str(row) in self.texts[str(column)]:
				self.texts[str(column)][str(row)].undraw()
				del self.texts[str(column)][str(row)]

		message = Text(Point(column, row), text.upper())
		message.setTextColor(self.currentTextColor)
		message.setSize(fontsize)
		message.draw(self.win)
		if str(row) not in self.texts:
			self.texts[str(column)] = {}
		self.texts[str(column)][str(row)] = message

	@_slow
	def pxlOn(self, row, column, color='blue'):
		# Row = y; Column = x
		pnt = Point(column, row)
		pnt.setOutline(self.tiColorToGraphicsColor(color))
		if column not in self.pixels:
			self.pixels[str(column)] = {}
		self.pixels[str(column)][str(row)] = pnt

		pnt.draw(self.win)

	@_slow
	def pxlOff(self, row, column):
		if str(column) in self.pixels:
			if str(row) in self.pixels[str(column)]:
				self.pixels[str(column)][str(row)].undraw()
				del self.pixels[str(column)][str(row)]

	@_slow
	def pxlTest(self, row, column):
		if str(column) in self.pixels and str(row) in self.pixels[str(column)]:
			return True
		return False

	@_slow
	def ptOn(self, x, y, mark=1, color='blue'):
		x, y = self.graphCoordsToPixels(x, y)

		if str(x) not in self.points:
			self.points[str(x)] = {}
		if str(y) not in self.points[str(x)]:
			self.points[str(x)][str(y)] = {}

		# If mark is unknown, it will default to 1, so test for 1 with `else`
		if mark in [2, 6]:
			# 3x3 box
			p1x = (x - 3 / 2)
			p1y = (y + 3 / 2)
			p2x = (x + 3 / 2)
			p2y = (y - 3 / 2)
			rec = Rectangle(Point(p1x, p1y), Point(p2x, p2y))
			rec.setOutline(self.tiColorToGraphicsColor(color))
			rec.draw(self.win)
			self.points[str(x)][str(y)][str(mark)] = (rec,)
		elif mark in [3, 7]:
			# 3x3 cross
			line1 = Line(Point(x - 2, y), Point(x + 2, y))
			line1.setOutline(self.tiColorToGraphicsColor(color))
			line1.draw(self.win)
			line2 = Line(Point(x, y - 2), Point(x, y + 2))
			line2.setOutline(self.tiColorToGraphicsColor(color))
			line2.draw(self.win)
			self.points[str(x)][str(y)][str(mark)] = (line1, line2)
		else:
			# Dot
			p1x = (x - 2 / 2)
			p1y = (y + 2 / 2)
			p2x = (x + 2 / 2)
			p2y = (y - 2 / 2)
			rec = Rectangle(Point(p1x, p1y), Point(p2x, p2y))
			rec.setFill(self.tiColorToGraphicsColor(color))
			rec.setOutline(self.tiColorToGraphicsColor(color))
			rec.draw(self.win)
			self.points[str(x)][str(y)][str(mark)] = (rec,)

	@_slow
	def ptOff(self, x, y, mark=1):
		x, y = self.graphCoordsToPixels(x, y)
		if str(x) in self.points:
			if str(y) in self.points[str(x)]:
				if str(mark) in self.points[str(x)][str(y)]:
					for item in self.points[str(x)][str(y)][str(mark)]:
						item.undraw()
					del self.points[str(x)][str(y)][str(mark)]



import re
import itertools
import tokenize
import io

try:
	from . import token_utils
except ImportError:
	import token_utils


def closeOpen(string):
	if string.count("\"") % 2 != 0:
		string = string + "\""

	# Split string by "
	splitString = string.split('"')
	open_paren = 0
	closed_paren = 0

	for i in range(0, len(splitString), 2): # skip all even elements since those are in between quotes
		open_paren += splitString[i].count('(')
		closed_paren += splitString[i].count(')')

	# Add needed closing parentheses
	string = string + ')' * (open_paren - closed_paren)
	return string


def fixEquals(string):
	# Change = to == if not between quotes
	return re.sub(r'(?!\B"[^"]*)(?<![<>!])\=(?!\()+(?![^"]*"\B)', " == ", string)


def noStringReplace(regexp, newSubstr, statementList):
	# Replace everything in statementList that matches regexp with newSubstr, as long as its not in a string
	# Split by "
	statementList = [item.split('"') for item in statementList]

	# Iterate through expression list
	for i in range(len(statementList)):
		# Skip all even elements since those are in between quotes
		for j in range(0, len(statementList[i]), 2):
			statementList[i][j] = re.sub(regexp, newSubstr, statementList[i][j])

		# Restore list element
		statementList[i] = '"'.join(statementList[i])

	return statementList


def parenthesis_split(sentence, separator=" ", lparen="(", rparen=")"):
	nb_brackets = 0
	sentence = sentence.strip(separator) # get rid of leading/trailing seps

	l = [0]
	for i, c in enumerate(sentence):
		if c == lparen:
			nb_brackets += 1
		elif c == rparen:
			nb_brackets -= 1
		elif c == separator and nb_brackets == 0:
			l.append(i)
		# handle malformed string
		if nb_brackets < 0:
			raise Exception("Syntax error")

	l.append(len(sentence))
	# handle missing closing parentheses
	if nb_brackets > 0:
		raise Exception("Syntax error")

	return([sentence[i:j].strip(separator) for i, j in zip(l, l[1:])])


def menu(title, args):
	choices = []
	for i in range(0, len(args), 2):
		choices.append((str((i // 2) + 1) + ".", str(args[i])))

	# Convert args to dict
	args = {str((i // 2) + 1) + ".": args[i + 1] for i in range(0, len(args), 2)}

	ifstmt = ["d = dialog.Dialog()", "menu = d.menu(\"" + title + "\", choices=" + str(choices) + ")"]
	for key in args:
		ifstmt.extend(["if menu[1] == \"" + str(key) + "\":", "\tclear()", "\tgoto .lbl" + str(args[key])])

	return ifstmt


def toValidEqn(source):
	"""This adds a multiplication symbol where it would be understood as
	being implicit by the normal way algebraic equations are written but would
	be a SyntaxError in Python. Thus we have::
		2N  -> 2*N
		N 2 -> N* 2
		2(A+B) -> 2*(A+B)
		(A+B)2 -> (A+B)*2
		2 3 -> 2* 3
		M N -> M* N
		(A+B)C -> (A+B)*C
		A(3) -> A*(3)
		a(3) -> a(3) - will only add multiplication if the preceding token is capital, since that is a variable
	"""

	"""
	Modified from ideas
	https://github.com/aroberge/ideas/blob/master/ideas/examples/implicit_multiplication.py
	"""

	constants = ['BLUE', 'RED', 'BLACK', 'MAGENTA', 'GREEN', 'ORANGE', 'BROWN', 'NAVY', 'LTBLUE', 'YELLOW', 'WHITE', 'LTGRAY', 'MEDGRAY', 'GRAY', 'DARKGRAY']

	tokens = token_utils.tokenize(source)
	if not tokens:
		return tokens

	prev_token = tokens[0]
	new_tokens = [prev_token]

	for token in tokens[1:]:
		if token.is_not_in(constants):
			# Check if implicit multiplication should be added
			if (
				(
					(prev_token.is_number() or (prev_token.is_identifier() and prev_token.string.isupper()))
					and ((token.is_identifier() and token.string.isupper()) or token.is_number() or token == "(")
				)
				or (
					(prev_token.is_identifier() and prev_token.string.isupper())
					and ((token.is_identifier() and token.string.isupper()) or token.is_number())
				)
				or (prev_token == ")" and ((token.is_identifier() and token.string.isupper()) or token.is_number()))
			):
				new_tokens.append("*")

			if token.is_identifier() and token.string.isupper() and len(token.string) > 1:
				# Multiple variables next to one another
				# ABC -> A*B*C
				token.string = '*'.join(token.string)
				new_tokens.append(token)
			else:
				new_tokens.append(token)
		else:
			# Token in constants, skip
			new_tokens.append(token)

		prev_token = token

	return token_utils.untokenize(new_tokens)


def decistmt(s):
	# Function for parsing a line of code and making all floats into decimal objects

	# List containing all tokens of new statement
	result = []

	# Tokenize statement
	tokens = tokenize.tokenize(io.BytesIO(s.encode('utf-8')).readline)

	# Iterate over tokens
	for toknum, tokval, _, _, _ in tokens:

		# Test if current token is a number and contains a '.' in order to detect floats
		if toknum == tokenize.NUMBER and '.' in tokval: # is float

			# Current token is float, so insert a new decimal object into the new statement instead of the current token
			# Example: 3.3 -> decimal.Decimal("3.3")
			result.extend([
				(tokenize.NAME, 'decimal'),
				(tokenize.OP, '.'),
				(tokenize.NAME, 'Decimal'),
				(tokenize.OP, '('),
				(tokenize.STRING, repr(tokval)),
				(tokenize.OP, ')')
			])
		else:
			# Current token is not a float, append it to the new statement without modification
			result.append((toknum, tokval))

	# Convert token list back into source code
	return tokenize.untokenize(result).decode('utf-8')

